/*

GooBer CPU core. 
Copyright (C) 2002 Ian Smith

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "gbcpu.h"
#include "cycles.h"

#define PREFIX_NA 0
#define PREFIX_CB 1
#define PREFIX_10 2

#define ZERO_FLAG  128
#define SUBT_FLAG  64
#define HCRY_FLAG  32
#define CARY_FLAG  16

#define WORD(hi, lo)   (lo | (hi << 8))
#define LO(a)          (a & 0xFF)
#define HI(a)          (a >> 8)

#define READ_B(ADDR)  (gbcpu->read)(ADDR)
  
#define READ_W(ADDR)  (READ_B(ADDR) | (READ_B(ADDR+1) << 8))

#define WRITE_B(ADDR, VALUE) (gbcpu->write)(ADDR, VALUE)
  
#define WRITE_W(ADDR, VALUE) WRITE_B(ADDR, LO(VALUE)); WRITE_B((ADDR)+1, HI(VALUE))

#define PUSH_B(VALUE) WRITE_B(--gbcpu->sp, VALUE)
#define PUSH_W(VALUE) PUSH_B(HI(VALUE)); PUSH_B(LO(VALUE))

#define POP_B()      READ_B(gbcpu->sp++)
#define POP_W(V)     V = (POP_B() | (POP_B() << 8))

#define A    gbcpu->a
#define B    gbcpu->b
#define C    gbcpu->c
#define D    gbcpu->d
#define E    gbcpu->e
#define H    gbcpu->h
#define L    gbcpu->l

#define AF   WORD(gbcpu->a, gbcpu->flags)
#define BC   WORD(gbcpu->b, gbcpu->c)
#define DE   WORD(gbcpu->d, gbcpu->e)
#define HL   WORD(gbcpu->h, gbcpu->l)
#define SP   gbcpu->sp
#define PC   gbcpu->pc

#define FLAGS gbcpu->flags

#define ZF   ((FLAGS & ZERO_FLAG) == ZERO_FLAG)
#define NF   ((FLAGS & SUBT_FLAG) == SUBT_FLAG)
#define HF   ((FLAGS & HCRY_FLAG) == HCRY_FLAG)
#define CF   ((FLAGS & CARY_FLAG) == CARY_FLAG)

#define set_AF(VALUE) gbcpu->a = HI(VALUE); gbcpu->flags = LO(VALUE)
#define set_BC(VALUE) gbcpu->b = HI(VALUE); gbcpu->c = LO(VALUE)
#define set_DE(VALUE) gbcpu->d = HI(VALUE); gbcpu->e = LO(VALUE)
#define set_HL(VALUE) gbcpu->h = HI(VALUE); gbcpu->l = LO(VALUE)

unsigned char bits[] = { 1, 2, 4, 8, 16, 32, 64, 128 };

#define FZ 128
#define FN 64
#define FH 32
#define FC 16

/* - Flag Macros - */
#define CL(FLAG) \
  if ((gbcpu->flags & FLAG) == FLAG){ gbcpu->flags ^= FLAG; }

#define ST(FLAG) gbcpu->flags |= FLAG

#define TGL(COND, FLAG) if (COND) { ST(FLAG); } else { CL(FLAG); }

#define IS_SET(FLAG) ((gbcpu->flags & FLAG) == FLAG)

#include "tables.h"
#include "mflags.h"

void gbcpu_reset(gbcpu_context *gbcpu)
{
  int i;

  gbcpu->pc = 0x100;
  gbcpu->sp = 0xFFFE;
  gbcpu->a = 0x01;
  gbcpu->flags = 0xB0;
  gbcpu->b = 0x00;
  gbcpu->c = 0x13;
  gbcpu->d = 0x00;
  gbcpu->e = 0xD8;
  gbcpu->h = 0x01;
  gbcpu->l = 0x4D;
  gbcpu->ime = 0;
  
  gbcpu->cycles = 0;
  
  gbcpu->cpu_mode = CPU_NORMAL;
}

int gbcpu_ex(gbcpu_context *gbcpu)
{
  unsigned short cycles, addr;
  unsigned char opcode, lsb, msb, bit;

  if (gbcpu->cpu_mode == CPU_NORMAL) {
    opcode = READ_B(gbcpu->pc);
    gbcpu->pc++;
    switch(opcode) {
    case 0x10:
      opcode = READ_B(gbcpu->pc);
      gbcpu->pc++;
      if (opcode != 0) {
	printf("GBCPU panic: [$%.4x] : ", gbcpu->pc);
	printf("unsupported or invalid opcode 0x10 %.2x\n",opcode);
	return 0;
      }
      gbcpu->cycles = 4;
      gbcpu->cpu_mode = CPU_HALT;
      break;

    case 0xCB:
      opcode = READ_B(gbcpu->pc);
      gbcpu->pc++;

      /* Execute BIT instruction */
      if ((opcode&0xC0)==0x40) {
	bit = (opcode&0x38)>>3;
	switch(opcode&0x07) {
	case 0x07: BIT(bit, A); break; /* BIT b, A */
	case 0x00: BIT(bit, B); break; /* BIT b, B */
	case 0x01: BIT(bit, C); break; /* BIT b, C */
	case 0x02: BIT(bit, D); break; /* BIT b, D */
	case 0x03: BIT(bit, E); break; /* BIT b, E */
	case 0x04: BIT(bit, H); break; /* BIT b, H */
	case 0x05: BIT(bit, L); break; /* BIT b, L */
	case 0x06: BIT(bit, READ_B(HL)); break; /* BIT b, (HL) */
	}
	break;
      }

      /* Execute RES instruction */
      if ((opcode&0xC0)==0x80) {
	bit = (opcode&0x38)>>3;
	switch(opcode&0x07) {
	case 0x07: RES(bit, A); break; /* RES b, A */
	case 0x00: RES(bit, B); break; /* RES b, B */
	case 0x01: RES(bit, C); break; /* RES b, C */
	case 0x02: RES(bit, D); break; /* RES b, D */
	case 0x03: RES(bit, E); break; /* RES b, E */
	case 0x04: RES(bit, H); break; /* RES b, H */
	case 0x05: RES(bit, L); break; /* RES b, L */
	case 0x06: lsb = READ_B(HL); RES(bit, lsb); WRITE_B(HL, lsb); break; /* RES b, (HL) */
	}
	break;
      }

      /* Execute SET instruction */
      if ((opcode&0xC0)==0xC0) {
	bit = (opcode&0x38)>>3;
	switch(opcode&0x07) {
	case 0x07: SET(bit, A); break; /* SET b, A */
	case 0x00: SET(bit, B); break; /* SET b, B */
	case 0x01: SET(bit, C); break; /* SET b, C */
	case 0x02: SET(bit, D); break; /* SET b, D */
	case 0x03: SET(bit, E); break; /* SET b, E */
	case 0x04: SET(bit, H); break; /* SET b, H */
	case 0x05: SET(bit, L); break; /* SET b, L */
	case 0x06: lsb = READ_B(HL); SET(bit, lsb); WRITE_B(HL, lsb); break; /* RES b, (HL) */
	}
	break;
      }

      switch (opcode) {
	
#include "cb_opcodes.h"

      default:
	printf("GBCPU panic: [$%.4x] : ", gbcpu->pc);
	printf("unsupported or invalid opcode 0xCB %.2x\n",opcode);
	return 0;
      }
      gbcpu->cycles = cb_cycles_table[opcode]*4;
      break;

#include "opcodes.h"

    default:
      printf("GBCPU panic: [$%.4x] : ", gbcpu->pc);
      printf("unsupported or invalid opcode %.2x\n",opcode);
      return 0;
    }
    gbcpu->cycles = cycles_table[opcode]*4;
    return 1;
  } 
  else {
    return 1;
  }
}

/* Set an interrupt */
int gbcpu_int(gbcpu_context *gbcpu, unsigned short addr)
{
  if (gbcpu->ime) {
    if (gbcpu->cpu_mode == CPU_HALT) {
      gbcpu->cpu_mode = CPU_NORMAL;
    }
    // push PC onto the stack
    PUSH_W(PC);
    PC = addr;

    return 1;
  }
  else {
    return 0;
  }
}

/* Disassemble current instruction */
void gbcpu_da(gbcpu_context *gbcpu, char out[])
{
  int offset = PC;
  unsigned char opcode;

  opcode = READ_B(offset++);
  switch (opcode) {
  case 0xCB:
    opcode = READ_B(offset++);
    switch(opcode) {
#include "d_cb_opcodes.h"
    default:
      sprintf(out, "<invalid opcode>\n");
      break;
    }
#include "d_opcodes.h"
  default:
    sprintf(out, "<invalid opcode>\n");
    break;
  }

}





