/*

GooBer Portable gameboy emulator
Copyright (C) 2002 Ian Smith

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include "mem.h"
#include "lcdc.h"

/* Pointers to memory banks */
unsigned char *rom0 = NULL;
unsigned char *rom1 = NULL;
unsigned char *vram = NULL;
unsigned char *exram = NULL;
unsigned char *wram0 = NULL;
unsigned char *wram1 = NULL;
unsigned char *oam = NULL;
unsigned char *hram = NULL;

/* Pointers to access handlers */
MEMREAD rom0_read_handler;
MEMWRITE rom0_write_handler;

MEMREAD rom1_read_handler;
MEMWRITE rom1_write_handler;

MEMREAD vram_read_handler;
MEMWRITE vram_write_handler;

MEMREAD oam_read_handler;
MEMWRITE oam_write_handler;

MEMREAD exram_read_handler;
MEMWRITE exram_write_handler;

MEMREAD wram1_read_handler;
MEMWRITE wram1_write_handler;

MEMREAD io_read_handler;
MEMWRITE io_write_handler;

void dump_bank(unsigned char *bank, char *fname, unsigned short sz)
{
  FILE *fd;
  fd = fopen(fname, "wb");
  fwrite(bank, 1, sz, fd);
  fclose(fd);
}

void memory_init()
{
}

void memory_reset()
{
  memset(vram, 0, 0x2000);
  memset(exram, 0, 0x2000);
  memset(wram0, 0, 0x1000);
  memset(wram1, 0, 0x1000);
  memset(oam, 0, 0x9F);
  memset(hram, 0, 0x7F);
}

void memory_cleanup()
{
  dump_bank(rom0, "rom0.dump", 0x4000);
  dump_bank(rom1, "rom1.dump", 0x4000);
  dump_bank(vram, "vram.dump", 0x2000);
  dump_bank(exram, "exram.dump", 0x2000);
  dump_bank(wram0, "wram0.dump", 0x1000);
  dump_bank(wram1, "wram1.dump", 0x1000);
  dump_bank(oam, "oam.dump", 0x9F);
  dump_bank(hram, "hram.dump", 0x7F);
}

/* Encapsulate all memory accesses in these two functions */
unsigned char mem_read(unsigned short a)
{
  /* ROM0 */
  if (a <= 0x3FFF)
    return rom0_read_handler(a);
  /* ROM1 */
  else if ((a >= 0x4000) && (a <= 0x7FFF))
    return rom1_read_handler(a-0x4000);
  /* VRAM */
  else if ((a >= 0x8000) && (a <= 0x9FFF))
/*     return (can_access_vram() ? vram_read_handler(a-0x8000) : 0xFF); */
    return vram_read_handler(a-0x8000);
  /* EXRAM */
  else if ((a >= 0xA000) && (a <= 0xBFFF))
    return exram_read_handler(a-0xA000);
  /* WRAM0 */
  else if ((a >= 0xC000) && (a <= 0xCFFF))
    return wram0[a-0xC000];
  /* WRAM1 */
  else if ((a >= 0xD000) && (a <= 0xDFFF))
    return wram1_read_handler(a-0xD000);
  /* ECHO */
  else if ((a >= 0xE000) && (a <= 0xFDFF)) {
    if ((a >= 0xE000) && (a <= 0xEFFF))
      return wram0[a-0xE000];
    else
      return wram1_read_handler(a-0xF000);
  }
  /* OAM */
  else if ((a >= 0xFE00) && (a <= 0xFE9F))
    return oam_read_handler(a-0xFE00);
  /* IO */
  else if ((a >= 0xFF00) && (a <= 0xFF7F))
    return io_read_handler(a-0xFF00);
  /* HRAM */
  else if ((a >= 0xFF80) && (a <= 0xFFFE))
    return hram[a-0xFF80];
  /* IE Register */
  else if (a == 0xFFFF) 
    return intr_read_reg(0xFF);
  /* Invalid address */
  else 
    return 0xFF;
}

void mem_write(unsigned short a, unsigned char b)
{
  /* ROM0 */
  if (a <= 0x3FFF) {
    rom0_write_handler(a, b);
  }
  /* ROM1 */
  else if ((a >= 0x4000) && (a <= 0x7FFF)) {
    rom1_write_handler(a-0x4000, b);
  }
  /* VRAM */
  else if ((a >= 0x8000) && (a <= 0x9FFF)) {
    if (can_access_vram()) vram_write_handler(a-0x8000, b);
  }
  /* EXRAM */
  else if ((a >= 0xA000) && (a <= 0xBFFF)) {
    exram_write_handler(a-0xA000, b);
  }
  /* WRAM0 */
  else if ((a >= 0xC000) && (a <= 0xCFFF)) {
    wram0[a-0xC000] = b;
  }
  /* WRAM1 */
  else if ((a >= 0xD000) && (a <= 0xDFFF)) {
    wram1_write_handler(a-0xD000, b); 
  }
  /* ECHO */
  else if ((a >= 0xE000) && (a <= 0xFDFF)) {
    if ((a >= 0xE000) && (a <= 0xEFFF))
      wram0[a-0xE000] = b;
    else
      wram1_write_handler(a-0xF000, b);
  }
  /* OAM */
  else if ((a >= 0xFE00) && (a <= 0xFE9F)) {
    oam_write_handler(a-0xFE00, b);
  }
  /* IO */
  else if ((a >= 0xFF00) && (a <= 0xFF7F)) {
    io_write_handler(a-0xFF00, b);
  }
  /* HRAM */
  else if ((a >= 0xFF80) && (a <= 0xFFFE)) {
    hram[a-0xFF80] = b;
  }
  /* IE Register */
  else if (a == 0xFFFF) {
    intr_write_reg(0xFF, b);
  }
    
}

/* Set source and handlers for all switchable banks */
void set_rom0_source(unsigned char *p)
{
  rom0 = p;
}

void set_rom0_handlers(MEMREAD r, MEMWRITE w)
{
  rom0_read_handler = r;
  rom0_write_handler = w;
}

void set_rom1_source(unsigned char *p)
{
  rom1 = p;
}

void set_rom1_handlers(MEMREAD r, MEMWRITE w)
{
  rom1_read_handler = r;
  rom1_write_handler = w;
}

void set_vram_source(unsigned char *p)
{
  vram = p;
}

void set_vram_handlers(MEMREAD r, MEMWRITE w)
{
  vram_read_handler = r;
  vram_write_handler = w;
}

void set_exram_source(unsigned char *p)
{
  exram = p;
}

void set_exram_handlers(MEMREAD r, MEMWRITE w)
{
  exram_read_handler = r;
  exram_write_handler = w;
}

void set_wram1_source(unsigned char *p)
{
  wram1 = p;
}

void set_wram1_handlers(MEMREAD r, MEMWRITE w)
{
  wram1_read_handler = r;
  wram1_write_handler = w;
}

void set_wram0_source(unsigned char *p)
{
  wram0 = p;
}

void set_oam_source(unsigned char *p)
{
  oam = p;
}

void set_oam_handlers(MEMREAD r, MEMWRITE w)
{
  oam_read_handler = r;
  oam_write_handler = w;
}

void set_hram_source(unsigned char *p)
{
  hram = p;
}

/* IO is special, it needs no source but it need handlers */
void set_io_handlers(MEMREAD r, MEMWRITE w)
{
  io_read_handler = r;
  io_write_handler = w;
}





